import pytest
import trio_asyncio
import asyncpg

import triopg


def unwrap(record):
    assert isinstance(record, asyncpg.Record)
    return tuple(record.values())


@pytest.mark.trio
async def test_connection_closed(asyncio_loop, postgresql_connection_specs):
    async with triopg.connect(**postgresql_connection_specs) as conn:
        pass
    with pytest.raises(triopg.InterfaceError):
        await conn.execute("VALUES (1)")


@pytest.mark.trio
async def test_cursor(triopg_conn):
    async with triopg_conn.transaction():
        cursor_factory = triopg_conn.cursor(
            "VALUES ($1, 1), ($2, 2), ($3, 3), ($4, 4), ($5, 5)", "1", "2",
            "3", "4", "5"
        )
        cursor = await cursor_factory

        row = await cursor.fetchrow()
        assert unwrap(row) == ("1", 1)
        advanced = await cursor.forward(1)
        assert advanced == 1
        fetched = await cursor.fetch(2)
        assert [unwrap(x) for x in fetched] == [("3", 3), ("4", 4)]

        items = []
        async for row in triopg_conn.cursor("VALUES ($1, 1), ($2, 2), ($3, 3)",
                                            "1", "2", "3"):
            items.append(unwrap(row))
        assert items == [("1", 1), ("2", 2), ("3", 3)]


@pytest.mark.trio
async def test_transaction(triopg_conn, asyncpg_execute):
    # Execute without transaction
    await triopg_conn.execute(
        """
        DROP TABLE IF EXISTS users;
        CREATE TABLE IF NOT EXISTS users (
            _id SERIAL PRIMARY KEY,
            user_id VARCHAR(32) UNIQUE
        )"""
    )
    assert await asyncpg_execute("SELECT * FROM users") == "SELECT 0"

    # Execute in transaction without exception
    async with triopg_conn.transaction():
        await triopg_conn.execute("INSERT INTO users (user_id) VALUES (1)")
    assert await asyncpg_execute("SELECT * FROM users") == "SELECT 1"

    # Execute in transaction raising exception, request should not be executed
    with pytest.raises(Exception):
        async with triopg_conn.transaction():
            await triopg_conn.execute("INSERT INTO users (user_id) VALUES (2)")
            raise Exception()

    assert await asyncpg_execute("SELECT * FROM users") == "SELECT 1"


@pytest.mark.trio
async def test_prepared_statement(triopg_conn):
    # Execute with prepared statement
    stmt = await triopg_conn.prepare("VALUES ($1, 2)")
    records = await stmt.fetchval("1")
    assert records == "1"

    # Test cursor in prepared statement
    async with triopg_conn.transaction():
        stmt = await triopg_conn.prepare(
            "VALUES ($1, 1), ($2, 2), ($3, 3), ($4, 4), ($5, 5)"
        )

        cursor = await stmt.cursor("1", "2", "3", "4", "5")
        row = await cursor.fetchrow()
        assert unwrap(row) == ("1", 1)
        advanced = await cursor.forward(1)
        assert advanced == 1
        fetched = await cursor.fetch(2)
        assert [unwrap(x) for x in fetched] == [("3", 3), ("4", 4)]

        items = []
        async for row in stmt.cursor("1", "2", "3", "4", "5"):
            items.append(unwrap(row))
        assert items == [("1", 1), ("2", 2), ("3", 3), ("4", 4), ("5", 5)]


@pytest.mark.trio
async def test_execute_many(triopg_conn, asyncpg_execute):
    await triopg_conn.execute(
        """
        DROP TABLE IF EXISTS users;
        CREATE TABLE IF NOT EXISTS users (
            _id SERIAL PRIMARY KEY,
            user_id VARCHAR(32) UNIQUE
        )"""
    )

    user_ids = [(str(i),) for i in range(10)]
    await triopg_conn.executemany(
        "INSERT INTO users (user_id) VALUES ($1)", user_ids
    )

    assert await asyncpg_execute("SELECT * FROM users") == "SELECT 10"


@pytest.mark.trio
async def test_use_pool_without_acquire_connection(triopg_pool):
    rep = await triopg_pool.execute(
        """
        DROP TABLE IF EXISTS users;
        CREATE TABLE IF NOT EXISTS users (
            _id SERIAL PRIMARY KEY,
            user_id VARCHAR(32) UNIQUE
        )"""
    )
    assert rep == "CREATE TABLE"

    user_ids = [(str(i),) for i in range(10)]
    await triopg_pool.executemany(
        "INSERT INTO users (user_id) VALUES ($1)", user_ids
    )

    rep = await triopg_pool.fetch("SELECT * FROM users")
    rep = [dict(x.items()) for x in rep]
    assert rep == [{"_id": i + 1, "user_id": str(i)} for i in range(10)]

    rep = await triopg_pool.fetchrow("SELECT * FROM users WHERE _id = $1", 1)
    assert dict(rep.items()) == {"_id": 1, "user_id": "0"}

    val = await triopg_pool.fetchval(
        "SELECT user_id FROM users WHERE _id = $1", 2
    )
    assert val == "1"



"""Top-level package for triopg."""

from ._version import __version__
from ._triopg import connect, create_pool
from .exceptions import *  # NOQA

__all__ = (
    '__version__',
    'connect',
    'create_pool',
) + exceptions.__all__  # NOQA



import pytest
import trio_asyncio
import asyncpg
from asyncpg.cluster import Cluster
import tempfile

import triopg


@pytest.fixture()
async def asyncio_loop():
    async with trio_asyncio.open_loop() as loop:
        yield loop


@pytest.fixture(scope='session')
def cluster():
    cluster_dir = tempfile.mkdtemp()
    cluster = Cluster(cluster_dir)

    cluster.init()
    try:
        cluster.start(port='dynamic')
        yield cluster
        cluster.stop()
    finally:
        cluster.destroy()


@pytest.fixture
def postgresql_connection_specs(cluster):
    return {'database': 'postgres', **cluster.get_connection_spec()}


@pytest.fixture()
async def asyncpg_conn(asyncio_loop, postgresql_connection_specs):
    conn = await trio_asyncio.aio_as_trio(asyncpg.connect
                                          )(**postgresql_connection_specs)
    try:
        yield conn
    finally:
        await trio_asyncio.aio_as_trio(conn.close)()


@pytest.fixture
def asyncpg_execute(asyncpg_conn):
    @trio_asyncio.aio_as_trio
    async def _asyncpg_execute(sql):
        return await asyncpg_conn.execute(sql)

    return _asyncpg_execute


@pytest.fixture(params=["from_connect", "from_pool"])
async def triopg_conn(request, asyncio_loop, postgresql_connection_specs):
    if request.param == "from_connect":
        async with triopg.connect(**postgresql_connection_specs) as conn:
            yield conn

    else:
        async with triopg.create_pool(**postgresql_connection_specs) as pool:
            async with pool.acquire() as conn:
                yield conn


@pytest.fixture
async def triopg_pool(asyncio_loop, postgresql_connection_specs):
    async with triopg.create_pool(**postgresql_connection_specs) as pool:
        yield pool



from functools import wraps, partial
import trio
import asyncpg
import trio_asyncio


def _shielded(f):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        with trio.CancelScope(shield=True):
            return await f(*args, **kwargs)

    return wrapper


def connect(*args, **kwargs):
    return TrioConnectionProxy(*args, **kwargs)


def create_pool(*args, **kwargs):
    return TrioPoolProxy(*args, **kwargs)


class TrioTransactionProxy:
    def __init__(self, asyncpg_transaction):
        self._asyncpg_transaction = asyncpg_transaction

    @trio_asyncio.aio_as_trio
    async def __aenter__(self, *args):
        return await self._asyncpg_transaction.__aenter__(*args)

    @_shielded
    @trio_asyncio.aio_as_trio
    async def __aexit__(self, *args):
        return await self._asyncpg_transaction.__aexit__(*args)


class TrioCursorProxy:
    def __init__(self, asyncpg_cursor):
        self._asyncpg_cursor = asyncpg_cursor

    @trio_asyncio.aio_as_trio
    async def fetch(self, *args, **kwargs):
        return await self._asyncpg_cursor.fetch(*args, **kwargs)

    @trio_asyncio.aio_as_trio
    async def fetchrow(self, *args, **kwargs):
        return await self._asyncpg_cursor.fetchrow(*args, **kwargs)

    @trio_asyncio.aio_as_trio
    async def forward(self, *args, **kwargs):
        return await self._asyncpg_cursor.forward(*args, **kwargs)


class TrioCursorFactoryProxy:
    def __init__(self, asyncpg_transaction_factory):
        self._asyncpg_transaction_factory = asyncpg_transaction_factory
        self._asyncpg_cursor_aiter = None

    def __await__(self):
        return self._wrapped_asyncpg_await().__await__()

    @trio_asyncio.aio_as_trio
    async def _wrapped_asyncpg_await(self):
        asyncpg_cursor = await self._asyncpg_transaction_factory
        return TrioCursorProxy(asyncpg_cursor)

    def __aiter__(self):
        self._asyncpg_cursor_aiter = self._asyncpg_transaction_factory.__aiter__(
        )
        return self

    @trio_asyncio.aio_as_trio
    async def __anext__(self):
        return await self._asyncpg_cursor_aiter.__anext__()


class TrioStatementProxy:
    def __init__(self, asyncpg_statement):
        self._asyncpg_statement = asyncpg_statement

    def cursor(self, *args, **kwargs):
        asyncpg_cursor_factory = self._asyncpg_statement.cursor(
            *args, **kwargs
        )
        return TrioCursorFactoryProxy(asyncpg_cursor_factory)

    def __getattr__(self, attr):
        target = getattr(self._asyncpg_statement, attr)

        if callable(target):

            @wraps(target)
            @trio_asyncio.aio_as_trio
            async def wrapper(*args, **kwargs):
                return await target(*args, **kwargs)

            # Only generate the function wrapper once per instance
            setattr(self, attr, wrapper)

            return wrapper

        return target


class TrioConnectionProxy:
    def __init__(self, *args, **kwargs):
        self._asyncpg_create_connection = partial(
            asyncpg.connect, *args, **kwargs
        )
        self._asyncpg_conn = None

    def transaction(self, *args, **kwargs):
        asyncpg_transaction = self._asyncpg_conn.transaction(*args, **kwargs)
        return TrioTransactionProxy(asyncpg_transaction)

    async def prepare(self, *args, **kwargs):
        asyncpg_statement = await trio_asyncio.aio_as_trio(
            self._asyncpg_conn.prepare(*args, **kwargs)
        )
        return TrioStatementProxy(asyncpg_statement)

    def __getattr__(self, attr):
        target = getattr(self._asyncpg_conn, attr)

        if callable(target):

            @wraps(target)
            @trio_asyncio.aio_as_trio
            async def wrapper(*args, **kwargs):
                return await target(*args, **kwargs)

            # Only generate the function wrapper once per connection instance
            setattr(self, attr, wrapper)

            return wrapper

        return target

    def cursor(self, *args, **kwargs):
        asyncpg_cursor_factory = self._asyncpg_conn.cursor(*args, **kwargs)
        return TrioCursorFactoryProxy(asyncpg_cursor_factory)

    @_shielded
    @trio_asyncio.aio_as_trio
    async def close(self):
        return await self._asyncpg_conn.close()

    async def __aenter__(self):
        if not self._asyncpg_conn:
            self._asyncpg_conn = await trio_asyncio.aio_as_trio(
                self._asyncpg_create_connection
            )()
        return self

    async def __aexit__(self, *exc):
        return await self.close()


class TrioPoolAcquireContextProxy:
    def __init__(self, asyncpg_acquire_context):
        self._asyncpg_acquire_context = asyncpg_acquire_context

    @trio_asyncio.aio_as_trio
    async def __aenter__(self, *args):
        proxy = await self._asyncpg_acquire_context.__aenter__(*args)
        conn_proxy = TrioConnectionProxy()
        conn_proxy._asyncpg_conn = proxy._con
        return conn_proxy

    @_shielded
    @trio_asyncio.aio_as_trio
    async def __aexit__(self, *args):
        return await self._asyncpg_acquire_context.__aexit__(*args)


class TrioPoolProxy:
    def __init__(self, *args, **kwargs):
        self._asyncpg_create_pool = partial(
            asyncpg.create_pool, *args, **kwargs
        )
        self._asyncpg_pool = None

    def acquire(self):
        return TrioPoolAcquireContextProxy(self._asyncpg_pool.acquire())

    async def execute(self, statement: str, *args, timeout: float = None):
        async with self.acquire() as conn:
            return await conn.execute(statement, *args, timeout=timeout)

    async def executemany(
            self, statement: str, args, *, timeout: float = None
    ):
        async with self.acquire() as conn:
            return await conn.executemany(statement, args, timeout=timeout)

    async def fetch(self, query, *args, timeout: float = None):
        async with self.acquire() as conn:
            return await conn.fetch(query, *args, timeout=timeout)

    async def fetchval(self, query, *args, timeout: float = None):
        async with self.acquire() as conn:
            return await conn.fetchval(query, *args, timeout=timeout)

    async def fetchrow(self, query, *args, timeout: float = None):
        async with self.acquire() as conn:
            return await conn.fetchrow(query, *args, timeout=timeout)

    @_shielded
    @trio_asyncio.aio_as_trio
    async def close(self):
        return await self._asyncpg_pool.close()

    def terminate(self):
        return self._asyncpg_pool.terminate()

    async def __aenter__(self):
        if not self._asyncpg_pool:
            self._asyncpg_pool = await trio_asyncio.aio_as_trio(
                self._asyncpg_create_pool
            )()
        return self

    async def __aexit__(self, *exc):
        return await self.close()
